@@ -13,24 +13,24 @@
 import argparse
 import logging
 import os.path
+import queue
 import shutil
-import threading
 from collections.abc import Callable, Iterator, Sequence
 from concurrent import futures
 from enum import Enum
 from pathlib import Path
-from typing import Literal
+from typing import Literal, TypeVar
 
 import beets
+import beets.plugins
+import beetsplug.convert as convert
 import confuse
 from beets import art, util
 from beets.library import Item, Library, parse_query_string
 from beets.plugins import BeetsPlugin
-from beets.ui import Subcommand, UserError, decargs, get_path_formats, input_yn, print_
+from beets.ui import Subcommand, UserError, get_path_formats, input_yn, print_
 from typing_extensions import Never, override
 
-import beetsplug.convert as convert
-
 
 class AlternativesPlugin(BeetsPlugin):
     def __init__(self):
@@ -57,8 +57,7 @@ class AlternativesPlugin(BeetsPlugin):
 
     def list_tracks(self, lib: Library, options: argparse.Namespace):
         if options.format is not None:
-            (fmt,) = decargs([options.format])
-            beets.config[Item._format_config_key].set(fmt)  # pyright: ignore[reportPrivateUsage]
+            beets.config[Item._format_config_key].set(options.format)  # pyright: ignore[reportPrivateUsage]
 
         alt = self.alternative(options.name, lib)
 
@@ -233,17 +232,29 @@ class Config:
 
 
 class Action(Enum):
-    ADD = 1
-    REMOVE = 2
-    WRITE = 3
-    MOVE = 4
-    SYNC_ART = 5
+    """Action to take for a track when syncing a collection"""
+
+    #: Track was not present in the collection before and is added
+    ADD = "ADD"
+
+    #: Remove the track from the collection
+    REMOVE = "REMOVE"
+
+    #: Write tags
+    WRITE = "WRITE"
+
+    #: Move the file for an existing track in a collection to a different path.
+    MOVE = "MOVE"
+
+    #: Write album art to the track’s metadata
+    SYNC_ART = "SYNC_ART"
 
 
 class External:
     def __init__(self, log: logging.Logger, lib: Library, config: Config):
         self._log = log
         self._config = config
+        self._embed = False
         self.lib = lib
         self.path_key = f"alt.{config.collection_id}"
         self.max_workers = int(str(beets.config["convert"]["threads"]))
@@ -312,47 +323,92 @@ class External:
         )
         return input_yn(msg, require=True)
 
-    def update(self, create: bool | None = None):
+    def update(self, create: bool | None = None):  # noqa: C901
         if not self._config.directory.is_dir() and not self.ask_create(create):
             print_(f"Skipping creation of {self._config.directory}")
             return
 
-        converter = self._converter()
-        for item, actions in self._items_actions():
-            dest = self.destination(item)
-            path = self._get_stored_path(item)
-            for action in actions:
-                if action == Action.MOVE:
-                    assert path is not None  # action guarantees that `path` is not none
-                    print_(f">{path} -> {dest}")
-                    dest.parent.mkdir(parents=True, exist_ok=True)
-                    path.rename(dest)
-                    # beets types are confusing
-                    util.prune_dirs(str(path.parent), root=str(self._config.directory))  # pyright: ignore
-                    self._set_stored_path(item, dest)
-                    item.store()
-                    path = dest
-                elif action == Action.WRITE:
-                    assert path is not None  # action guarantees that `path` is not none
-                    print_(f"*{path}")
-                    item.write(path=bytes(path))
-                elif action == Action.SYNC_ART:
-                    print_(f"~{path}")
-                    assert path is not None
-                    self._sync_art(item, path)
-                elif action == Action.ADD:
-                    print_(f"+{dest}")
-                    converter.run(item)
-                elif action == Action.REMOVE:
-                    assert path is not None  # action guarantees that `path` is not none
-                    print_(f"-{path}")
-                    self._remove_file(item)
-                    item.store()
-
-        for item, dest in converter.as_completed():
+        def finalize_converted_item(item: Item, dest: Path):
+            # Don't rely on the converter to write correct/complete tags.
+            item.write(path=bytes(dest))
+            if self._embed:
+                self._sync_art(item, dest)
             self._set_stored_path(item, dest)
             item.store()
-        converter.shutdown()
+            _send_item_updated(
+                collection=self._config.collection_id,
+                path=dest,
+                item=item,
+                action=action,
+            )
+
+        converter, converting_done = self._converter()
+        with converter as converter:
+            for item, actions in self._items_actions():
+                dest = self.destination(item)
+                path = self._get_stored_path(item)
+                for action in actions:
+                    delay_finalize = False
+                    if action == Action.MOVE:
+                        assert (
+                            path is not None
+                        )  # action guarantees that `path` is not none
+                        print_(f">{path} -> {dest}")
+                        dest.parent.mkdir(parents=True, exist_ok=True)
+                        path.rename(dest)
+                        # beets types are confusing
+                        util.prune_dirs(
+                            str(path.parent), root=str(self._config.directory)
+                        )  # pyright: ignore
+                        self._set_stored_path(item, dest)
+                        item.store()
+                        path = dest
+                    elif action == Action.WRITE:
+                        assert (
+                            path is not None
+                        )  # action guarantees that `path` is not none
+                        print_(f"*{path}")
+                        item.write(path=bytes(path))
+                    elif action == Action.SYNC_ART:
+                        print_(f"~{path}")
+                        assert path is not None
+                        self._sync_art(item, path)
+                    elif action == Action.ADD:
+                        print_(f"+{dest}")
+                        dest.parent.mkdir(exist_ok=True, parents=True)
+                        if self._should_transcode(item):
+                            delay_finalize = True
+                            converter.run(item, dest)
+                        else:
+                            self._log.debug(f"copying {dest}")
+                            shutil.copyfile(item.path, dest)
+                            if self._embed:
+                                self._sync_art(item, dest)
+                            self._set_stored_path(item, dest)
+                            item.store()
+
+                    elif action == Action.REMOVE:
+                        assert (
+                            path is not None
+                        )  # action guarantees that `path` is not none
+                        print_(f"-{path}")
+                        self._remove_file(item)
+                        item.store()
+
+                    if not delay_finalize:
+                        _send_item_updated(
+                            collection=self._config.collection_id,
+                            path=dest,
+                            item=item,
+                            action=action,
+                        )
+
+                for item, dest in _get_queue_available(converting_done):
+                    finalize_converted_item(item, dest)
+
+            for item, dest in converter.as_completed():
+                for item, dest in _get_queue_available(converting_done):
+                    finalize_converted_item(item, dest)
 
     def destination(self, item: Item) -> Path:
         """Returns the path for `item` in the external collection."""
@@ -385,14 +441,15 @@ class External:
             util.prune_dirs(str(path), root=str(self._config.directory))  # pyright: ignore
         del item[self.path_key]
 
-    def _converter(self) -> "Worker":
-        def _convert(item: Item):
-            dest = self.destination(item)
-            dest.parent.mkdir(exist_ok=True, parents=True)
-            shutil.copyfile(item.path, dest)
-            return item, dest
+    def _converter(self) -> tuple["Worker", queue.Queue[tuple[Item, Path]]]:
+        done_queue = queue.Queue()
 
-        return Worker(_convert, self.max_workers)
+        def _convert(item: Item, dest: Path):
+            raise RuntimeError(
+                "Convert must never be called for non-converting collection"
+            )
+
+        return (Worker(_convert, self.max_workers), done_queue)
 
     def _sync_art(self, item: Item, path: Path):
         """Embed artwork in the file at `path`."""
@@ -408,6 +465,9 @@ class External:
                 itempath=bytes(path),
             )
 
+    def _should_transcode(self, item: Item) -> bool:
+        return False
+
 
 class ExternalConvert(External):
     def __init__(
@@ -424,26 +484,15 @@ class ExternalConvert(External):
         self.convert_cmd, self.ext = convert.get_format(self._formats[0])
 
     @override
-    def _converter(self) -> "Worker":
-        fs_lock = threading.Lock()
+    def _converter(self) -> tuple["Worker", queue.Queue[tuple[Item, Path]]]:
+        done_queue = queue.Queue()
 
-        def _convert(item: Item):
-            dest = self.destination(item)
-            with fs_lock:
-                dest.parent.mkdir(exist_ok=True, parents=True)
-
-            if self._should_transcode(item):
-                self._encode(self.convert_cmd, item.path, bytes(dest))
-                # Don't rely on the converter to write correct/complete tags.
-                item.write(path=bytes(dest))
-            else:
-                self._log.debug(f"copying {dest}")
-                shutil.copyfile(item.path, dest)
-            if self._embed:
-                self._sync_art(item, dest)
+        def _convert(item: Item, dest: Path):
+            self._encode(self.convert_cmd, item.path, bytes(dest))
+            done_queue.put((item, dest))
             return item, dest
 
-        return Worker(_convert, self.max_workers)
+        return Worker(_convert, self.max_workers), done_queue
 
     @override
     def destination(self, item: Item) -> Path:
@@ -453,6 +502,7 @@ class ExternalConvert(External):
         else:
             return dest
 
+    @override
     def _should_transcode(self, item: Item):
         return item.format.lower() not in self._formats
 
@@ -492,17 +542,24 @@ class SymlinkView(External):
                     self._remove_file(item)
                     self._create_symlink(item)
                     self._set_stored_path(item, dest)
+                    item.store()
                 elif action == Action.ADD:
                     print_(f"+{dest}")
                     self._create_symlink(item)
                     self._set_stored_path(item, dest)
+                    item.store()
                 elif action == Action.REMOVE:
                     assert path is not None  # action guarantees that `path` is not none
                     print_(f"-{path}")
                     self._remove_file(item)
-                else:
-                    continue
-                item.store()
+                    item.store()
+
+                _send_item_updated(
+                    collection=self._config.collection_id,
+                    path=dest,
+                    item=item,
+                    action=action,
+                )
 
     def _create_symlink(self, item: Item):
         dest = self.destination(item)
@@ -522,23 +579,45 @@ class SymlinkView(External):
 
 class Worker(futures.ThreadPoolExecutor):
     def __init__(
-        self, fn: Callable[[Item], tuple[Item, Path]], max_workers: int | None
+        self, fn: Callable[[Item, Path], tuple[Item, Path]], max_workers: int | None
     ):
         super().__init__(max_workers)
         self._tasks: set[futures.Future[tuple[Item, Path]]] = set()
         self._fn = fn
 
-    def run(self, item: Item):
-        fut = self.submit(self._fn, item)
+    def run(self, item: Item, path: Path):
+        fut = self.submit(self._fn, item, path)
         self._tasks.add(fut)
         return fut
 
-    def as_completed(self):
-        for f in futures.as_completed(self._tasks):
+    def as_completed(self, timeout: float | None = None):
+        for f in futures.as_completed(self._tasks, timeout=timeout):
             self._tasks.remove(f)
             yield f.result()
 
 
+_T = TypeVar("_T")
+
+
+def _get_queue_available(q: queue.Queue[_T] | queue.SimpleQueue[_T]):
+    while True:
+        try:
+            item = q.get(block=False)
+        except queue.Empty:
+            break
+        yield item
+
+
+def _send_item_updated(*, collection: str, path: Path, item: Item, action: Action):
+    beets.plugins.send(
+        "alternatives.item_updated",
+        collection=collection,
+        path=path,
+        item=item,
+        action=action.value,
+    )
+
+
 _beets_version = tuple(map(int, beets.__version__.split(".")[0:2]))
 
 if _beets_version >= (2, 3):
